import { CommandClasses, Duration, IVirtualEndpoint, IZWaveEndpoint, IZWaveNode, Maybe, NODE_ID_BROADCAST, SendCommandOptions, SupervisionResult, TXReport, ValueChangeOptions, ValueDB, ValueID } from "@zwave-js/core";
import type { ZWaveApplicationHost } from "@zwave-js/host";
import { OnlyMethods } from "@zwave-js/shared";
export declare type ValueIDProperties = Pick<ValueID, "property" | "propertyKey">;
/** Used to identify the method on the CC API class that handles setting values on nodes directly */
export declare const SET_VALUE: unique symbol;
export declare type SetValueImplementation = (property: ValueIDProperties, value: unknown, options?: SetValueAPIOptions) => Promise<SupervisionResult | undefined>;
/**
 * A generic options bag for the `setValue` API.
 * Each implementation will choose the options that are relevant for it, so you can use the same options everywhere.
 * @publicAPI
 */
export declare type SetValueAPIOptions = Partial<ValueChangeOptions>;
/** Used to identify the method on the CC API class that handles polling values from nodes */
export declare const POLL_VALUE: unique symbol;
export declare type PollValueImplementation<T = unknown> = (property: ValueIDProperties) => Promise<T | undefined>;
export declare function throwUnsupportedProperty(cc: CommandClasses, property: string | number): never;
export declare function throwUnsupportedPropertyKey(cc: CommandClasses, property: string | number, propertyKey: string | number): never;
export declare function throwMissingPropertyKey(cc: CommandClasses, property: string | number): never;
export declare function throwWrongValueType(cc: CommandClasses, property: string | number, expectedType: string, receivedType: string): never;
export interface SchedulePollOptions {
    duration?: Duration;
    transition?: "fast" | "slow";
}
/**
 * The base class for all CC APIs exposed via `Node.commandClasses.<CCName>`
 * @publicAPI
 */
export declare class CCAPI {
    protected readonly applHost: ZWaveApplicationHost;
    protected readonly endpoint: IZWaveEndpoint | IVirtualEndpoint;
    constructor(applHost: ZWaveApplicationHost, endpoint: IZWaveEndpoint | IVirtualEndpoint);
    static create<T extends CommandClasses>(ccId: T, applHost: ZWaveApplicationHost, endpoint: IZWaveEndpoint | IVirtualEndpoint, requireSupport?: boolean): CommandClasses extends T ? CCAPI : CCToAPI<T>;
    /**
     * The identifier of the Command Class this API is for
     */
    readonly ccId: CommandClasses;
    protected [SET_VALUE]: SetValueImplementation | undefined;
    /**
     * Can be used on supported CC APIs to set a CC value by property name (and optionally the property key)
     */
    get setValue(): SetValueImplementation | undefined;
    /** Whether a successful setValue call should imply that the value was successfully updated */
    isSetValueOptimistic(valueId: ValueID): boolean;
    protected [POLL_VALUE]: PollValueImplementation | undefined;
    /**
     * Can be used on supported CC APIs to poll a CC value by property name (and optionally the property key)
     */
    get pollValue(): PollValueImplementation | undefined;
    /**
     * Schedules a value to be polled after a given time. Schedules are deduplicated on a per-property basis.
     * @returns `true` if the poll was scheduled, `false` otherwise
     */
    protected schedulePoll(property: ValueIDProperties, expectedValue: unknown, { duration, transition }?: SchedulePollOptions): boolean;
    /**
     * Retrieves the version of the given CommandClass this endpoint implements
     */
    get version(): number;
    /** Determines if this simplified API instance may be used. */
    isSupported(): boolean;
    /**
     * Determine whether the linked node supports a specific command of this command class.
     * "unknown" means that the information has not been received yet
     */
    supportsCommand(command: number): Maybe<boolean>;
    protected assertSupportsCommand(commandEnum: unknown, command: number): void;
    protected assertPhysicalEndpoint(endpoint: IZWaveEndpoint | IVirtualEndpoint): asserts endpoint is IZWaveEndpoint;
    /** Returns the command options to use for sendCommand calls */
    protected get commandOptions(): SendCommandOptions;
    /** Creates an instance of this API, scoped to use the given options */
    withOptions(options: SendCommandOptions): this;
    /** Creates an instance of this API which (if supported) will return TX reports along with the result. */
    withTXReport<T extends this>(): WithTXReport<T>;
    protected isSinglecast(): this is this & {
        endpoint: IZWaveEndpoint;
    };
    protected isMulticast(): this is this & {
        endpoint: IVirtualEndpoint & {
            nodeId: number[];
        };
    };
    protected isBroadcast(): this is this & {
        endpoint: IVirtualEndpoint & {
            nodeId: typeof NODE_ID_BROADCAST;
        };
    };
    /**
     * Returns the node this CC API is linked to. Throws if the controller is not yet ready.
     */
    getNode(): IZWaveNode | undefined;
    /** Returns the value DB for this CC API's node (if it can be safely accessed) */
    protected tryGetValueDB(): ValueDB | undefined;
    /** Returns the value DB for this CC's node (or throws if it cannot be accessed) */
    protected getValueDB(): ValueDB;
}
/** A CC API that is only available for physical endpoints */
export declare class PhysicalCCAPI extends CCAPI {
    constructor(applHost: ZWaveApplicationHost, endpoint: IZWaveEndpoint | IVirtualEndpoint);
    protected readonly endpoint: IZWaveEndpoint;
}
export declare type APIConstructor<T extends CCAPI = CCAPI> = new (applHost: ZWaveApplicationHost, endpoint: IZWaveEndpoint | IVirtualEndpoint) => T;
export declare type CCToName<CC extends CommandClasses> = [CC] extends [
    typeof CommandClasses["Alarm Sensor"]
] ? "Alarm Sensor" : [CC] extends [typeof CommandClasses["Association"]] ? "Association" : [CC] extends [typeof CommandClasses["Association Group Information"]] ? "Association Group Information" : [CC] extends [typeof CommandClasses["Barrier Operator"]] ? "Barrier Operator" : [CC] extends [typeof CommandClasses["Basic"]] ? "Basic" : [CC] extends [typeof CommandClasses["Battery"]] ? "Battery" : [CC] extends [typeof CommandClasses["Binary Sensor"]] ? "Binary Sensor" : [CC] extends [typeof CommandClasses["Binary Switch"]] ? "Binary Switch" : [CC] extends [typeof CommandClasses["CRC-16 Encapsulation"]] ? "CRC-16 Encapsulation" : [CC] extends [typeof CommandClasses["Central Scene"]] ? "Central Scene" : [CC] extends [typeof CommandClasses["Climate Control Schedule"]] ? "Climate Control Schedule" : [CC] extends [typeof CommandClasses["Clock"]] ? "Clock" : [CC] extends [typeof CommandClasses["Color Switch"]] ? "Color Switch" : [CC] extends [typeof CommandClasses["Configuration"]] ? "Configuration" : [CC] extends [typeof CommandClasses["Door Lock"]] ? "Door Lock" : [CC] extends [typeof CommandClasses["Door Lock Logging"]] ? "Door Lock Logging" : [CC] extends [typeof CommandClasses["Entry Control"]] ? "Entry Control" : [CC] extends [typeof CommandClasses["Firmware Update Meta Data"]] ? "Firmware Update Meta Data" : [CC] extends [typeof CommandClasses["Humidity Control Mode"]] ? "Humidity Control Mode" : [CC] extends [typeof CommandClasses["Humidity Control Operating State"]] ? "Humidity Control Operating State" : [CC] extends [typeof CommandClasses["Humidity Control Setpoint"]] ? "Humidity Control Setpoint" : [CC] extends [typeof CommandClasses["Indicator"]] ? "Indicator" : [CC] extends [typeof CommandClasses["Irrigation"]] ? "Irrigation" : [CC] extends [typeof CommandClasses["Language"]] ? "Language" : [CC] extends [typeof CommandClasses["Lock"]] ? "Lock" : [CC] extends [typeof CommandClasses["Manufacturer Proprietary"]] ? "Manufacturer Proprietary" : [CC] extends [typeof CommandClasses["Manufacturer Specific"]] ? "Manufacturer Specific" : [CC] extends [typeof CommandClasses["Meter"]] ? "Meter" : [CC] extends [typeof CommandClasses["Multi Channel Association"]] ? "Multi Channel Association" : [CC] extends [typeof CommandClasses["Multi Channel"]] ? "Multi Channel" : [CC] extends [typeof CommandClasses["Multi Command"]] ? "Multi Command" : [CC] extends [typeof CommandClasses["Multilevel Sensor"]] ? "Multilevel Sensor" : [CC] extends [typeof CommandClasses["Multilevel Switch"]] ? "Multilevel Switch" : [CC] extends [typeof CommandClasses["No Operation"]] ? "No Operation" : [CC] extends [typeof CommandClasses["Node Naming and Location"]] ? "Node Naming and Location" : [CC] extends [typeof CommandClasses["Notification"]] ? "Notification" : [CC] extends [typeof CommandClasses["Powerlevel"]] ? "Powerlevel" : [CC] extends [typeof CommandClasses["Protection"]] ? "Protection" : [CC] extends [typeof CommandClasses["Scene Activation"]] ? "Scene Activation" : [CC] extends [typeof CommandClasses["Scene Actuator Configuration"]] ? "Scene Actuator Configuration" : [CC] extends [typeof CommandClasses["Scene Controller Configuration"]] ? "Scene Controller Configuration" : [CC] extends [typeof CommandClasses["Schedule Entry Lock"]] ? "Schedule Entry Lock" : [CC] extends [typeof CommandClasses["Security 2"]] ? "Security 2" : [CC] extends [typeof CommandClasses["Security"]] ? "Security" : [CC] extends [typeof CommandClasses["Sound Switch"]] ? "Sound Switch" : [CC] extends [typeof CommandClasses["Supervision"]] ? "Supervision" : [CC] extends [typeof CommandClasses["Thermostat Fan Mode"]] ? "Thermostat Fan Mode" : [CC] extends [typeof CommandClasses["Thermostat Fan State"]] ? "Thermostat Fan State" : [CC] extends [typeof CommandClasses["Thermostat Mode"]] ? "Thermostat Mode" : [CC] extends [typeof CommandClasses["Thermostat Operating State"]] ? "Thermostat Operating State" : [CC] extends [typeof CommandClasses["Thermostat Setback"]] ? "Thermostat Setback" : [CC] extends [typeof CommandClasses["Thermostat Setpoint"]] ? "Thermostat Setpoint" : [CC] extends [typeof CommandClasses["Time"]] ? "Time" : [CC] extends [typeof CommandClasses["Time Parameters"]] ? "Time Parameters" : [CC] extends [typeof CommandClasses["User Code"]] ? "User Code" : [CC] extends [typeof CommandClasses["Version"]] ? "Version" : [CC] extends [typeof CommandClasses["Wake Up"]] ? "Wake Up" : [CC] extends [typeof CommandClasses["Z-Wave Plus Info"]] ? "Z-Wave Plus Info" : never;
export declare type CCToAPI<CC extends CommandClasses> = CCToName<CC> extends keyof CCAPIs ? CCAPIs[CCToName<CC>] : never;
export declare type APIMethodsOf<CC extends CommandClasses> = Omit<OnlyMethods<CCToAPI<CC>>, "isSetValueOptimistic" | "isSupported" | "supportsCommand" | "withOptions" | "withTXReport">;
export declare type OwnMethodsOf<API extends CCAPI> = Omit<OnlyMethods<API>, keyof OnlyMethods<CCAPI>>;
export declare type WrapWithTXReport<T> = [T] extends [Promise<infer U>] ? Promise<WrapWithTXReport<U>> : [T] extends [void] ? {
    txReport: TXReport | undefined;
} : {
    result: T;
    txReport: TXReport | undefined;
};
export declare type WithTXReport<API extends CCAPI> = Omit<API, keyof OwnMethodsOf<API> | "withOptions" | "withTXReport"> & {
    [K in keyof OwnMethodsOf<API>]: API[K] extends (...args: any[]) => any ? (...args: Parameters<API[K]>) => WrapWithTXReport<ReturnType<API[K]>> : never;
};
export interface CCAPIs {
    [Symbol.iterator](): Iterator<CCAPI>;
    "Alarm Sensor": import("../cc/AlarmSensorCC").AlarmSensorCCAPI;
    Association: import("../cc/AssociationCC").AssociationCCAPI;
    "Association Group Information": import("../cc/AssociationGroupInfoCC").AssociationGroupInfoCCAPI;
    "Barrier Operator": import("../cc/BarrierOperatorCC").BarrierOperatorCCAPI;
    Basic: import("../cc/BasicCC").BasicCCAPI;
    Battery: import("../cc/BatteryCC").BatteryCCAPI;
    "Binary Sensor": import("../cc/BinarySensorCC").BinarySensorCCAPI;
    "Binary Switch": import("../cc/BinarySwitchCC").BinarySwitchCCAPI;
    "CRC-16 Encapsulation": import("../cc/CRC16CC").CRC16CCAPI;
    "Central Scene": import("../cc/CentralSceneCC").CentralSceneCCAPI;
    "Climate Control Schedule": import("../cc/ClimateControlScheduleCC").ClimateControlScheduleCCAPI;
    Clock: import("../cc/ClockCC").ClockCCAPI;
    "Color Switch": import("../cc/ColorSwitchCC").ColorSwitchCCAPI;
    Configuration: import("../cc/ConfigurationCC").ConfigurationCCAPI;
    "Door Lock": import("../cc/DoorLockCC").DoorLockCCAPI;
    "Door Lock Logging": import("../cc/DoorLockLoggingCC").DoorLockLoggingCCAPI;
    "Entry Control": import("../cc/EntryControlCC").EntryControlCCAPI;
    "Firmware Update Meta Data": import("../cc/FirmwareUpdateMetaDataCC").FirmwareUpdateMetaDataCCAPI;
    "Humidity Control Mode": import("../cc/HumidityControlModeCC").HumidityControlModeCCAPI;
    "Humidity Control Operating State": import("../cc/HumidityControlOperatingStateCC").HumidityControlOperatingStateCCAPI;
    "Humidity Control Setpoint": import("../cc/HumidityControlSetpointCC").HumidityControlSetpointCCAPI;
    Indicator: import("../cc/IndicatorCC").IndicatorCCAPI;
    Irrigation: import("../cc/IrrigationCC").IrrigationCCAPI;
    Language: import("../cc/LanguageCC").LanguageCCAPI;
    Lock: import("../cc/LockCC").LockCCAPI;
    "Manufacturer Proprietary": import("../cc/ManufacturerProprietaryCC").ManufacturerProprietaryCCAPI;
    "Manufacturer Specific": import("../cc/ManufacturerSpecificCC").ManufacturerSpecificCCAPI;
    Meter: import("../cc/MeterCC").MeterCCAPI;
    "Multi Channel Association": import("../cc/MultiChannelAssociationCC").MultiChannelAssociationCCAPI;
    "Multi Channel": import("../cc/MultiChannelCC").MultiChannelCCAPI;
    "Multi Command": import("../cc/MultiCommandCC").MultiCommandCCAPI;
    "Multilevel Sensor": import("../cc/MultilevelSensorCC").MultilevelSensorCCAPI;
    "Multilevel Switch": import("../cc/MultilevelSwitchCC").MultilevelSwitchCCAPI;
    "No Operation": import("../cc/NoOperationCC").NoOperationCCAPI;
    "Node Naming and Location": import("../cc/NodeNamingCC").NodeNamingAndLocationCCAPI;
    Notification: import("../cc/NotificationCC").NotificationCCAPI;
    Powerlevel: import("../cc/PowerlevelCC").PowerlevelCCAPI;
    Protection: import("../cc/ProtectionCC").ProtectionCCAPI;
    "Scene Activation": import("../cc/SceneActivationCC").SceneActivationCCAPI;
    "Scene Actuator Configuration": import("../cc/SceneActuatorConfigurationCC").SceneActuatorConfigurationCCAPI;
    "Scene Controller Configuration": import("../cc/SceneControllerConfigurationCC").SceneControllerConfigurationCCAPI;
    "Schedule Entry Lock": import("../cc/ScheduleEntryLockCC").ScheduleEntryLockCCAPI;
    "Security 2": import("../cc/Security2CC").Security2CCAPI;
    Security: import("../cc/SecurityCC").SecurityCCAPI;
    "Sound Switch": import("../cc/SoundSwitchCC").SoundSwitchCCAPI;
    Supervision: import("../cc/SupervisionCC").SupervisionCCAPI;
    "Thermostat Fan Mode": import("../cc/ThermostatFanModeCC").ThermostatFanModeCCAPI;
    "Thermostat Fan State": import("../cc/ThermostatFanStateCC").ThermostatFanStateCCAPI;
    "Thermostat Mode": import("../cc/ThermostatModeCC").ThermostatModeCCAPI;
    "Thermostat Operating State": import("../cc/ThermostatOperatingStateCC").ThermostatOperatingStateCCAPI;
    "Thermostat Setback": import("../cc/ThermostatSetbackCC").ThermostatSetbackCCAPI;
    "Thermostat Setpoint": import("../cc/ThermostatSetpointCC").ThermostatSetpointCCAPI;
    Time: import("../cc/TimeCC").TimeCCAPI;
    "Time Parameters": import("../cc/TimeParametersCC").TimeParametersCCAPI;
    "User Code": import("../cc/UserCodeCC").UserCodeCCAPI;
    Version: import("../cc/VersionCC").VersionCCAPI;
    "Wake Up": import("../cc/WakeUpCC").WakeUpCCAPI;
    "Z-Wave Plus Info": import("../cc/ZWavePlusCC").ZWavePlusCCAPI;
}
//# sourceMappingURL=API.d.ts.map